/* 

WARNING: DO NOT COPY THIS FILE TO OTHER PROJECTS. IT HAS BEEN
MODIFIED FOR THE CMP3 PROJECT. SPECIFICALLY, THE INPUT STREAM
OF THE CHILD PROCESS IS DUPED TO A TEMPORARY FILE INSTEAD OF 
A PIPE. */

#include <errno.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "proc.h"

typedef struct pr {
	struct pr *prev,*next;
	int pid;
	int status;
	int fds[6];
	int socks;

	int ecode; /* exit reason (exit/signal) */
	int earg;  /* exit code or signal name */
} pr_t;

static pr_t head= { &head, &head };

#define PROC_RUNNING	1
#define PROC_DEAD		2
#define PROC_STOPPED	3


static int sig_installed=0;
static struct sigaction  old_sa;

	
/*
	one-shot execution: 
		execute the program with the given input string, collect output and error streams and return
	monitored executon:
		start the program
		enable the user to read/write the output/error/input streams
		enable the user to kill/stop the program
		let the user know whether the program is still alive or not
*/


int proc_kill(int id,int sig) {
	if (kill(id,sig))
		return -errno;
	return 0;
}



static void close_force(int fd) {
	if (fd!=-1)
		while(close(fd) && errno==EINTR)
			;
}


static void close_sockets(pr_t *p) {
	int i;
	for(i=0;i<6;i++) {
		close_force(p->fds[i]);
		p->fds[i]= -1;
	}
}


/* FIXME: in case of errors, put a descriptive text into the error stream
          mini-shell is a good name for this purpose
*/


static void child(pr_t *p,char *fn,char **argv,int opts) {
	int ret= 127;

#if 0
	if (opts & PROC_INP) {
		while (dup2(p->fds[0],0)<0)
			if (errno!=EINTR)
				goto duperr;
		close_force(p->fds[1]);
	}
#endif
	ret--;

	if (opts & PROC_OUTP) {
		while (dup2(p->fds[3],1)<0)
			if (errno!=EINTR)
				goto duperr;
		close_force(p->fds[2]);
	}
	ret--;

	if (opts & PROC_ERR)  {
		while (dup2(p->fds[5],2)<0)
			if (errno!=EINTR)
				goto duperr;
		close_force(p->fds[4]);
	}
	ret--;

	execvp(fn,argv);
duperr:
	exit(ret);
}

static pr_t *find(int pid) {
	pr_t *p;
	for(p=head.next;p!=&head;p=p->next)
		if (p->pid==pid)
			return p;
	return NULL;
}

static char *code_to_str(int code) {
	switch(code) {
		case CLD_EXITED: return "exit";
		case CLD_KILLED: return "killed";
		case CLD_DUMPED: return "dumped";
	}
	return "unknown";
}
static void child_sig(int sno,siginfo_t *si,void *dummy) {
	int pid= si->si_pid;
	int status= si->si_status;
	int code= si->si_code;
	pr_t *p;
	p= find(pid);
	if (!p)
		return ;
	switch(code) {
		case CLD_EXITED:
		case CLD_KILLED:
		case CLD_DUMPED:
			p->status= PROC_DEAD;
			p->ecode= code;
			p->earg= status;
#ifdef proc_self_test
			printf("child exit: code= %d (%s) arg= %d\n",code,code_to_str(code),status);
			fflush(0);
#endif
			break;
		case CLD_TRAPPED:
		case CLD_STOPPED:
			p->status= PROC_STOPPED;
			break;
		case CLD_CONTINUED:
			p->status= PROC_RUNNING;
			break;
	}
}

static void install_handler() {
	struct sigaction sa;
	if (sig_installed)
		return ;
	sa.sa_sigaction= child_sig;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags= SA_SIGINFO ;
	sa.sa_restorer= 0;
	while(sigaction(SIGCHLD, &sa, &old_sa) && errno==EINTR)
		;
	sig_installed= 1;
}


static void remove_node(pr_t *p) {
	p->prev->next= p->next;
	p->next->prev= p->prev;
	free(p);
}


static void close_socks_main(pr_t *p) {
	// close_force(p->fds[0]);
	close_force(p->fds[3]);
	close_force(p->fds[5]);
	p->fds[0]= -1;
	p->fds[3]= -1;
	p->fds[5]= -1;
}


int proc_start(char *fn,char **argv,int opts, int inp_fd) {
	
	int eno;
	pr_t *p;

	int i;
	
	install_handler();
	
	p= malloc(sizeof(*p));
	if (!p)
		return -ENOMEM;
	p->prev= head.prev;
	p->next= &head;
	head.prev->next= p;
	head.prev= p;
	p->status= PROC_RUNNING;

	for(i=0;i<6;i++)
		p->fds[i]= -1;

	if (opts & PROC_INP) {
#if 0
		if (pipe(p->fds+0))
			goto sockerr;
#else
		dup2(inp_fd, 0);
#endif
	}

	if (opts & PROC_OUTP) {
		if (pipe(p->fds+2))
			goto sockerr;
	}

	if (opts & PROC_ERR) {
		if (pipe(p->fds+4))
			goto sockerr;
	}

	p->pid= fork();
	if (p->pid<0)
		goto sockerr;
	if (p->pid==0)
		child(p,fn,argv,opts);
	if (0) {
		int i;
		for(i=0;i<6;i++)
			printf("fd[%d]= %d\n",i,p->fds[i]);
	}
	close_socks_main(p);
	return p->pid;

sockerr:
	eno= errno;
	close_sockets(p);
	remove_node(p);
	return -eno;
}

static void interp_status(int s,pr_t *p) {

	if (WIFEXITED(s)) {
		p->ecode= CLD_EXITED;
		p->earg= WEXITSTATUS(s);
	} else if (WIFSIGNALED(s)) {
		p->ecode= CLD_DUMPED;
		p->earg= WTERMSIG(s);
	}
	p->status= PROC_DEAD;
#ifdef proc_self_test
	printf("child exit (wait): code= %d (%s) arg= %d\n",p->ecode,code_to_str(p->ecode),p->earg);
#endif
}


int proc_end(int id,int *st) {
	int status;
	pr_t *p;
	int c,a;
	p= find(id);
	if (!p)
		return 0;
	close_sockets(p);
	while(waitpid(id,&status,0)<0 && errno==EINTR)
		;
	if (p->status!=PROC_DEAD)
		interp_status(status,p);
	c= p->ecode;
	a= p->earg;
	remove_node(p);
	if (c==CLD_KILLED || c==CLD_DUMPED || (c==CLD_EXITED && a!=0)) {
		if (st)
			*st= a;
		return c;
	}
	return 0;
}





int proc_input(int id) {
	pr_t *p= find(id);
	return p ? p->fds[1] : -ESRCH;
}


int proc_output(int id) {
	pr_t *p= find(id);
	return p ? p->fds[2] : -ESRCH;
}


int proc_error(int id) {
	pr_t *p= find(id);
	return p ? p->fds[4] : -ESRCH;
}


int proc_islive(int id) {
	pr_t *p= find(id);
	return p ? (p->status!=PROC_DEAD) : 0;
}





